@@ -4,8 +4,8 @@ import numpy as np
 from PIL import Image
 import io
 from typing import List
-from projectkiwi.tools import getOverlap, splitZXY
-from projectkiwi.models import Annotation
+from projectkiwi.tools import getOverlap, splitZXY, urlFromZxy
+from projectkiwi.models import Annotation, Project, ImageryLayer, TilePath, Task
 import threading
 import queue
 
@@ -13,66 +13,45 @@ import queue
 
 
 class Connector():
-    def __init__(self, key, url="https://project-kiwi.org/api/"):
+    def __init__(self, key, url="https://project-kiwi.org/"):
         """constructor
 
         Args:
             key (str): API key.
             url (str, optional): url for api, in case of multiple instances. Defaults to "https://project-kiwi.org/api/".
-
-        Raises:
-            ValueError: API key must be supplied.
-        """        
-        if key is None:
-            raise ValueError("API key missing")
+        """
 
         self.key = key
         self.url = url
 
 
-    def getImagery(self, project=None):
-        """Get a list of imagery
+    def getImagery(self, project_id: str) -> List[ImageryLayer]:
+        """Get a list of imagery layers for a project
 
         Args:
-            project (str, optional): ID of the project to get all the imagery for, by default, all projects associated with user.
+            project (str): ID of the project to get all the imagery for.
 
         Returns:
-            json: list of imagery
+            List[ImageryLayer]: list of imagery layers
         """        
-        route = "get_imagery"
-        params = {'key': self.key}
-        if not project is None:
-            params['project'] = project
-        else:
-            projects = self.getProjects()
-            assert len(projects) > 0, "No projects found"
-            assert len(projects) == 1, "Multiple projects found, please specify a single project"
-            params['project'] = projects[0]
+        
+        route = "api/get_imagery"
+        params = {
+            'key': self.key, 
+            'project': project_id
+        }
 
         r = requests.get(self.url + route, params=params)
         r.raise_for_status()
-        jsonResponse = r.json()
-        return jsonResponse
+        imageryList = r.json()
+        imagery = []
+        for layer in imageryList:
+            imagery.append(ImageryLayer(**layer))
+        assert len(imageryList) == len(imagery), "Failed to parse imagery"
+        return imagery
 
 
-    def getTiles(self, imageryId: str):
-        """Get a list of tiles for a given imagery id
-
-        Args:
-            imageryId (str): ID of the imagery to retrieve a list of tiles for
-
-        Returns:
-            list: list of tile urls
-        """        
-        route = "get_tile_list"
-        params = {'key': self.key, 'imagery_id': imageryId}
-
-        r = requests.get(self.url + route, params=params)
-        r.raise_for_status()
-        jsonResponse = r.json()
-        return jsonResponse
-
-    def getTile(self, url):
+    def readTile(self, url) -> np.ndarray:
         """Get a tile in numpy array form
 
         Args:
@@ -85,113 +64,138 @@ class Connector():
         r.raise_for_status()
         tileContent = r.content
         return np.asarray(Image.open(io.BytesIO(tileContent)))
+    
+    def getTile(self, 
+            z: int,
+            x: int,
+            y: int,
+            imagery_id: str
+        ) -> np.ndarray:
+        """Download a tile given the z,x,y and id
+
+        Args:
+            z (int): zoom
+            x (int): x tile
+            y (int): y tile
+            imagery_id (str): id of the imagery
+
+        Returns:
+            np.ndarray: numpy array of tile
+        """        
+
+        url = urlFromZxy(z, x, y, imagery_id, self.url, self.key)
+        
+        return self.readTile(url)
 
 
-    def getTileDict(self, imageryId: str):
-        """Get a dictionary of tiles for a given imagery id
+
+    def getTileList(self, 
+            imagery_id: str, 
+            zoom: int) -> List[TilePath]:
+        """Get a list of tiles for a given imagery id
 
         Args:
             imageryId (str): ID of the imagery to retrieve a list of tiles for
+            zoom (int): Zoom level
 
         Returns:
-            dict: a dictionary of tiles with zxy keys
-        """        
-        route = "get_tile_list"
-        params = {'key': self.key, 'imagery_id': imageryId}
+            List[TilePath]: A list of tiles with zxy and url
+        """
+        route = "api/get_tile_list"
+        params = {
+            'key': self.key, 
+            'imagery_id': imagery_id, 
+            'zoom': zoom}
 
         r = requests.get(self.url + route, params=params)
         r.raise_for_status()
-        jsonResponse = r.json()
-        dict = {}
-        for tile in jsonResponse:
-            dict[tile['zxy']] = tile['url']
-        return dict
+        tileList = r.json()
+        tiles = []
+        for tile in tileList:
+            tiles.append(TilePath(**tile))
+        assert len(tiles) == len(tileList), "Failed to parse tiles"
+        return tiles
 
 
-
-    def getImageryStatus(self, imageryId: str):
+    def getImageryStatus(self, imagery_id: str) -> str:
         """ Get the status of imagery
 
         Args:
-            imageryId (str): Imagery id
+            imagery_id (str): Imagery id
 
         Returns:
             str: status
         """        
-        route = "get_imagery_status"
-        params = {'key': self.key, 'imagery_id': imageryId}
+        route = "api/get_imagery_status"
+        params = {'key': self.key, 'imagery_id': imagery_id}
 
         r = requests.get(self.url + route, params=params)
         r.raise_for_status()
         return r.json()['status']
 
 
-    def setImageryStatus(self, imageryId: str, status: str):
-        """Set the status for imagery e.g. when the upload is complete
-
-        Args:
-            imageryId (str): Imagery id
-            status (str): status e.g. "awaiting processing"
-        """
-        route = "set_imagery_status" 
-        params = {'key': self.key, 'imagery_id': imageryId, 'status': status}
-
-        r = requests.get(self.url + route, params=params)
-        r.raise_for_status()
-
-    def getProjects(self):
-        """Get a list of projects you have access to
+    def getProjects(self) -> List[Project]:
+        """Get a list of projects for a user
 
         Returns:
-            List: projects
+            List[Projects]: projects
         """
-        route = "get_projects" 
+        route = "api/get_projects" 
         params = {'key': self.key}
 
         r = requests.get(self.url + route, params=params)
         r.raise_for_status()
 
         try:
-            projects = r.json()['projects']
-            assert len(projects) > 0, "Error: No projects found"
+            projectList = r.json()
+            assert len(projectList) > 0, "Error: No projects found"
+            projects = []
+            for proj in projectList:
+                projects.append(Project(**proj))
+            assert len(projectList) == len(projects), \
+                    f"Error: Could not parse projects, {projectList}"
             return projects
         except Exception as e:
-            print("Error: Could not find projects")
+            print("Error: Could not get projects")
             raise e
         
 
-    def addImagery(self, filename: str, name: str):
+    def addImagery(self, filename: str, name: str, project_id: str) -> str:
         """ Add imagery to project-kiwi.org
 
         Args:
             filename (str): Path to the file to be uploaded
             name (str): Name for the imagery
+            project_id (str): Id of the project to add the layer to
 
         Returns:
-            int: imagery id
+            str: imagery id
         """       
         
         # get presigned upload url
-        route = "get_imagery_upload_url"
-        params = {'key': self.key, 'filename': filename, 'name': name}
+        route = "api/get_imagery_upload_url"
+        params = {
+            'key': self.key, 
+            'filename': filename, 
+            'name': name,
+            'project_id': project_id
+        }
         r = requests.get(self.url + route, params=params)
         r.raise_for_status()
         jsonResponse = r.json()
         url = jsonResponse['url']
         
         # upload
-        r = requests.put(url, data=open(filename, 'rb'), headers={'Content-type': ''})
-        r.raise_for_status()
+        with open(filename, 'rb') as data:
+            r = requests.put(url, data=data, headers={'Content-type': ''})
+            r.raise_for_status()
 
-        # set the status - triggers processing
-        self.setImageryStatus(jsonResponse['imagery_id'], "awaiting processing")
         return jsonResponse['imagery_id']
     
 
     def getSuperTile(self, 
             zxy: str,
-            url: str = None,
-            imagery_id: str = None,
+            url: str,
             max_zoom: int = 22,
             num_threads: int = 4
     ) -> np.ndarray:
@@ -199,8 +203,7 @@ class Connector():
 
         Args:
             zxy (str): zxy for tile e.g. 12/345/678
-            url (str, optional): url template e.g. https://tile.openstreetmap.org/${z}/${x}/${y}.png . Defaults to None.
-            imagery_id (str, optional): imagery id for project kiwi managed imagery. Defaults to None.
+            url (str): url template e.g. https://tile.openstreetmap.org/${z}/${x}/${y}.png
             max_zoom (int, optional): full resolution tiles to use. Defaults to 22.
             num_threads (int, optional): number of threads for downloading. Defaults to 4.
 
@@ -218,7 +221,7 @@ class Connector():
             while True:
                 try:
                     i, j, url = q1.get(timeout=10)
-                    tile = self.getTile(url)
+                    tile = self.readTile(url)
                     q2.put((i,j,tile))
                     q1.task_done()
                 except queue.Empty as e:
@@ -229,14 +232,6 @@ class Connector():
 
 
         z,x,y = splitZXY(zxy)
-
-        # make urls
-        if url is None:
-            assert not imagery_id is None, "Please specify either an imagery id or url"
-            urlTemplate = "https://project-kiwi-tiles.s3.amazonaws.com/" + imagery_id + "/{z}/{x}/{y}"
-        else:
-            urlTemplate = url
-
         tile_width = 2**(max_zoom - z)
         width = 256*tile_width
         assert width < 10000, "Resultant image would be too large (100MP limit), try reducing max zoom or increasing super tile zoom"
@@ -251,10 +246,12 @@ class Connector():
                 x_prime = x*tile_width + i
                 y_prime = (y+1)*tile_width - (j+1)
 
-                imgUrl = urlTemplate
+                imgUrl = url
+                imgUrl = imgUrl.replace("{s}.", "")
                 imgUrl = imgUrl.replace("{z}", str(z_prime))
                 imgUrl = imgUrl.replace("{x}", str(x_prime))
                 imgUrl = imgUrl.replace("{y}", str(y_prime))
+                imgUrl = imgUrl.replace("{key}", self.key)
 
                 q1.put((i, j, imgUrl))
         
@@ -287,22 +284,20 @@ class Connector():
             return returnImg[:,:,:3]
 
     
-    def getAnnotations(self, project=None):
+    def getAnnotations(self, project_id: str) -> List[Annotation]:
         """Get all annotations in a project
+        Args:
+            project_id (str): id for the project to get the predictions for
 
         Returns:
             List[Annotation]: annotations
         """
 
-        route = "get_annotations" 
-        params = {'key': self.key, }
-        if not project is None:
-            params['project'] = project
-        else:
-            projects = self.getProjects()
-            assert len(projects) > 0, "No projects found"
-            assert len(projects) == 1, "Multiple projects found, please specify a single project"
-            params['project'] = projects[0]
+        route = "api/get_annotations"
+        params = {
+            'key': self.key,
+            'project': project_id
+        }
 
         r = requests.get(self.url + route, params=params)
         r.raise_for_status()
@@ -310,9 +305,9 @@ class Connector():
         try:
             annotations = []
             annotationsDict = r.json()
-            for id, data in annotationsDict.items():
-                annotations.append(Annotation.from_dict(data, id))
-
+            for annotation_id, data in annotationsDict.items():
+                annotations.append(Annotation.from_dict(data, annotation_id))
+            assert len(annotationsDict) == len(annotations), "ERROR: could not parse annotations"
             return annotations
 
         except Exception as e:
@@ -320,14 +315,15 @@ class Connector():
             raise e
 
 
-    def getPredictions(self, project=None):
+    def getPredictions(self, project_id: str) -> List[Annotation]:
         """Get all predictions in a project
-
+        Args:
+            project_id (str): id for the project to get the predictions for
         Returns:
             List[Annotation]: predictions
         """
         
-        annotations = self.getAnnotations(project=project)
+        annotations = self.getAnnotations(project_id=project_id)
 
         return [annotation for annotation in annotations if annotation.confidence != None]
 
@@ -342,7 +338,7 @@ class Connector():
         """ Filter a set of annotations for those that have overlap with some tile
 
         Args:
-            annotations (List[Annotation]): All annotations
+            annotations (List[Annotation]): Annotations to filter
             zxy (str): The tile e.g. 12/345/678
             overlap_threshold (float, optional): How much overlap. Defaults to 0.2.
 
@@ -362,23 +358,76 @@ class Connector():
 
         return annotationsInTile
 
-    def getTasks(self, queue_id: int) -> List[dict]:
+    def getTasks(self, queue_id: int) -> List[Task]:
         """Get a list of tasks in a queue.
 
         Args:
             queue_id (int): The ID of the queue
 
         Returns:
-            List[dict]: list of tasks, each is a dict
+            List[Task]: list of tasks
         """        
        
-        route = "get_tasks"
+        route = "api/get_tasks"
         params = {'key': self.key, "queue_id": queue_id}
 
         r = requests.get(self.url + route, params=params)
         r.raise_for_status()
-        jsonResponse = r.json()
-        return jsonResponse['task']
+        data = r.json()
+        assert data['success'] == True, "Failed to get tasks"
+        
+        taskList = data['task']
+        tasks = []
+        for task in taskList:
+            tasks.append(Task(**task))
+        
+        assert len(tasks) == len(taskList), "Failed to parse tasks"
+
+        return tasks
+    
+    def getTask(self, queue_id: int) -> List[Task]:
+        """Get a random task for a queue.
+
+        Args:
+            queue_id (int): The ID of the queue
+
+        Returns:
+            Task: task
+        """        
+       
+        route = "api/get_task"
+        params = {'key': self.key, "queue_id": queue_id}
+
+        r = requests.get(self.url + route, params=params)
+        r.raise_for_status()
+        data = r.json()
+        assert data['success'] == True, "Failed to get tasks"
+        
+        task = data['task']
+
+        return Task(**task)
+
+    def getNextTask(self, queue_id: int) -> List[Task]:
+        """Get a predictable next task for a queue.
+
+        Args:
+            queue_id (int): The ID of the queue
+
+        Returns:
+            Task: task
+        """        
+       
+        route = "api/get_next_task"
+        params = {'key': self.key, "queue_id": queue_id}
+
+        r = requests.get(self.url + route, params=params)
+        r.raise_for_status()
+        data = r.json()
+        assert data['success'] == True, "Failed to get tasks"
+        
+        task = data['task']
+
+        return Task(**task)
 
 
     def addAnnotation(self, annotation: Annotation, project: str) -> int:
@@ -391,7 +440,7 @@ class Connector():
         Returns:
             int: annotation id if successful
         """        
-        route = "add_annotation"
+        route = "api/add_annotation"
         headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
 
         annoDict = dict(annotation)
@@ -415,7 +464,7 @@ class Connector():
         """       
 
         assert not annotation.confidence is None, "No confidence for prediction"
-        route = "add_prediction"
+        route = "api/add_prediction"
         headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
 
         annoDict = dict(annotation)
@@ -424,4 +473,33 @@ class Connector():
         r = requests.post(self.url + route, data=json.dumps(annoDict), headers=headers)
         r.raise_for_status()
         jsonResponse = r.json()
-        return jsonResponse['annotation_id']
\ No newline at end of file
+        return jsonResponse['annotation_id']
+
+    def getImageryUrl(self, imagery_id: str, project_id: str) -> str:
+        """Get the url for imagery from it's id
+
+        Args:
+            imagery_id (str): Id for the imagery
+            project_id (str): Project to look in
+
+        Returns:
+            str: The url template
+        """        
+        imagery = self.getImagery(project_id)
+        imagery_url = [image.url for image in imagery if image.id == imagery_id][0]
+        return imagery_url
+        
+    def removeAllPredictions(self, project_id: str):
+        """Remove all predictions in a project
+
+        Args:
+            project_id (str): project id
+        """    
+
+        route = "api/remove_all_predictions" 
+        params = {'key': self.key, 'project': project_id}
+
+        r = requests.delete(self.url + route, 
+                headers={'Content-Type': 'application/json'},
+                data=json.dumps(params))
+        r.raise_for_status()
